@@ -55,19 +55,13 @@ use warnings;
 use Scalar::Util qw(refaddr blessed);
 use File::Basename;
 use IO::File;
-use File::Temp 0.17;
 use Mail::Log::Exceptions;
 use base qw(Exporter);
 
-
 BEGIN {
     use Exporter ();
-    use vars qw($VERSION @EXPORT @EXPORT_OK %EXPORT_TAGS @ISA);
-    $VERSION     = '1.1.0';
-    #Give a hoot don't pollute, do not export more than needed by default
-    @EXPORT      = qw();
-    @EXPORT_OK   = qw();
-    %EXPORT_TAGS = ();
+    use vars qw($VERSION);
+    $VERSION     = '1.0300_1';
 }
 
 #
@@ -75,6 +69,11 @@ BEGIN {
 #
 
 my %log_info;
+my %parse_buffer;
+my %parse_buffer_start_line;
+my %parse_buffer_size;
+my %debug;
+my %current_line;
 
 #
 # DESTROY class variables.
@@ -84,7 +83,14 @@ my %log_info;
 sub DESTROY {
 	my ($self) = @_;
 	
+	$log_info{refaddr $self}{filehandle}->close() if defined($log_info{refaddr $self}{filehandle});
+	
 	delete $log_info{refaddr $self};
+	delete $parse_buffer{refaddr $self};
+	delete $parse_buffer_start_line{refaddr $self};
+	delete $parse_buffer_size{refaddr $self};
+	delete $debug{refaddr $self};
+	delete $current_line{refaddr $self};
 	
 	return;
 }
@@ -98,14 +104,14 @@ use overload (
 	qw{""} => sub { my ($self) = @_;
 					return  blessed($self)
 							.' File: '
-							.$log_info{refaddr $self}{'filename'}
+							.$log_info{$$self}{'filename'}
 							.' Line: '
-							.$log_info{refaddr $self}{'current_line'};
+							.$current_line{$$self};
 					},
 	
 	# Boolean overloads to if we are usable.  (Have a filehandle.)
 	qw{bool} => sub { my ($self) = @_;
-						return defined($log_info{refaddr $self}{'filehandle'});
+						return defined($log_info{$$self}{'filehandle'});
 					},
 	
 	# Numeric context just doesn't mean anything.  Throw an error.
@@ -113,7 +119,7 @@ use overload (
 				},
 	
 	# Heh.  Iterator context is the same as 'next'...
-	q{<>} => sub { my ($self) = @_; return $self->next(); },
+	q{<>} => sub { return $_[0]->next(); },
 	
 	# Perl standard for everything else.
 	fallback => 1,
@@ -131,6 +137,12 @@ Example:
 Note that it is an error to call any method other than C<set_logfile> if you
 have not passed it in the constructor.
 
+Optional keys in the hash are 'buffer_length' and 'debug'.  The buffer length
+is the number of lines to read at a time (and store in the internal buffer).
+Default is 128.  Setting debug to a true value will result is some debugging
+information being printed to STDERR.  (I reserve the right to remove or change
+the debug info at any time.)
+
 =cut
 
 sub new
@@ -138,12 +150,24 @@ sub new
     my ($class, $parameters_ref) = @_;
 
     my $self = bless \do{my $anon}, $class;
+	$$self = refaddr $self;
 
 	# Log info.
 	if ( defined($parameters_ref->{'log_file'}) ) {
 		$self->set_logfile($parameters_ref->{'log_file'});  # Better to keep validation together.
 	}
 
+	$debug{$$self} = defined($parameters_ref->{debug});
+
+#$debug{refaddr $self} = 1;
+
+	# Init the buffer.
+	$current_line{$$self} = 0;
+	$parse_buffer_start_line{$$self} = 0;
+	$parse_buffer{$$self} = undef;
+	$parse_buffer_size{$$self} = defined($parameters_ref->{buffer_length}) ? $parameters_ref->{buffer_length} : 128;
+
+
 	return $self;
 }
 
@@ -155,9 +179,10 @@ success.
 
 Files can be compressed or uncompressed: If they are compressed, then
 C<IO::Uncompress::AnyUncompress> must be installed with the relevant
-decompression libraries.  Currently only 'tgz', 'zip', 'gz', and 'bz2' archives
-are supported, but there is no technical reason not to support more.  (It just
-keeps a couple of lines of code shorter.)
+decompression libraries. (As well as version 0.17 or better of File::Temp.)
+Currently only 'tgz', 'zip', 'gz', and 'bz2' archives are supported, but
+there is no technical reason not to support more.  (It just keeps a couple
+of lines of code shorter.)
 
 Note that to support seeking in the file the log will be uncompressed to disk
 before it is read: If there is insufficent space to do so, we may have trouble.
@@ -172,11 +197,11 @@ Example:
 sub set_logfile {
 	my ($self, $new_name) = @_;
 	
-	# Check to make sure the file exitsts,
+	# Check to make sure the file exists,
 	# and then that we can read it, before accpeting the filename.
 	if ( -e $new_name ) {
 		if ( -r $new_name ) {
-			$log_info{refaddr $self}{'filename'} = $new_name;
+			$log_info{$$self}{'filename'} = $new_name;
 
 			# We'll check the extension to see if it is compressed.
 			my (undef, undef, $suffix) = fileparse($new_name, qw(tgz zip gz bz2));
@@ -184,27 +209,32 @@ sub set_logfile {
 
 				# Since we only need uncompress symantics right here, we'll
 				# only load them if we need them.  Neat, huh?
-				require IO::Uncompress::AnyUncompress;
+				eval { require IO::Uncompress::AnyUncompress } or Mail::Log::Exceptions->throw("Need IO::Uncompress::AnyUncompress for compressed files.\n");
 				IO::Uncompress::AnyUncompress->import( qw(anyuncompress) );
 
+				# Same with File::Temp;
+				eval { require File::Temp } or Mail::Log::Exceptions->throw("Need File::Temp version 0.17 or better for compressed files.\n");
+				File::Temp->VERSION( 0.17 )	# Minimum version check.
+					or Mail::Log::Exceptions->throw("Need File::Temp version 0.17 or better for compressed files.\n");
+
 				# If it is compressed, uncompress to a temp file and use that.
 				my $temp = new File::Temp();
 				anyuncompress($new_name, $temp)
 					or Mail::Log::Exceptions::LogFile->throw("Unable to uncompress logfile $new_name: ". $IO::Uncompress::AnyUncompress::AnyUncompressError ."\n");
 				$temp->seek(0,0)
 					or Mail::Log::Exceptions::LogFile->throw("Unable to seek to beginning of temp file.\n");
-				$log_info{refaddr $self}{'filehandle'} = $temp;
+				$log_info{$$self}{'filehandle'} = $temp;
 			}
 			else {
 				# If it wasn't compressed, open it direct.
-				$log_info{refaddr $self}{'filehandle'} = IO::File->new($new_name, '<')
+				$log_info{$$self}{'filehandle'} = IO::File->new($new_name, '<')
 					or Mail::Log::Exceptions::LogFile->throw("Unable to open file $new_name: $!\n");
 			}
 
 			# Init some location information on the file.
-			$log_info{refaddr $self}->{'current_line'} = 0;
-			delete $log_info{refaddr $self}{'line_positions'};
-			${$log_info{refaddr $self}{'line_positions'}}[$log_info{refaddr $self}{'current_line'}] = $log_info{refaddr $self}{'filehandle'}->getpos();
+			$current_line{$$self} = 0;
+			delete $log_info{$$self}->{'line_positions'};
+			${$log_info{$$self}{'line_positions'}}[$current_line{$$self}] = $log_info{$$self}{'filehandle'}->getpos();
 		}
 		else {
 			Mail::Log::Exceptions::LogFile->throw("Log file $new_name is not readable.\n");
@@ -219,7 +249,7 @@ sub set_logfile {
 
 =head2 next
 
-Returns a reference to a hash of the next parsable line of the log, or undef on 
+Returns a reference to a hash of the next parsable line of the log, or 'undef' on 
 end of file/failure.
 
 There are a couple of required keys that any parser must implement:
@@ -233,11 +263,12 @@ program, and C<text> is the text following any 'standard' headers.  (Usually,
 minus those already required keys.)
 
 This version is just a placeholder: It will return a 
-'Mail::Log::Exceptions::Unimplemented' exception if called.  It is expected to
-be overridden by the subclass.  (And is the only method needed to be overridden.)
+'Mail::Log::Exceptions::Unimplemented' exception if called.  Subclasses are
+expected to override the C<_parse_next_line> method to get an operable parser.
+(And that is the only method needed to be overridden for a working subclass.)
 
 Other 'standard' fields that are expected in a certain format (but are not
-required to always be present) are 'from', 'to', 'size', 'subject'.  'to'
+required to always be present) are 'from', 'to', 'size', 'subject', delay.  'to'
 should point to an array of addresses.  (As listed in the log.  That includes 
 angle brackets, usually.)
 
@@ -258,7 +289,62 @@ or...
 sub next {
 	my ($self) = @_;
 	
-	Mail::Log::Exceptions::Unimplemented->throw("Method 'next' needs to be implemented by the subclass.\n");
+	# This is the same as $self->get_current_line();
+	# (Or at least it should be.  Done for speed.)
+	my $current_line = $current_line{$$self};
+	
+	if ( defined($parse_buffer{$$self})
+			and ( ($current_line+1) <= ($parse_buffer_start_line{$$self} + $#{$parse_buffer{$$self}}) )
+			and ( ($current_line+1) >= $parse_buffer_start_line{$$self})
+		) {
+		
+		# Increment where we are.
+		$current_line{$$self} = $current_line{$$self} + 1;
+
+#		print STDERR 'Returning line number '. $self->get_line_number() ." from buffer.\n" if $debug{$selfref};
+
+		# Return the data we were asked for.
+		return $parse_buffer{$$self}->[($current_line - $parse_buffer_start_line{$$self}+1)];
+	}
+	else {
+		# Move the actual read postition to where we are.
+		# (But only if we've acutally ever read anything.)
+		if ( defined($log_info{$$self}->{line_positions}->[$current_line]) ) {
+			$log_info{$$self}{filehandle}->setpos($log_info{$$self}->{line_positions}->[$current_line])
+				or Mail::Log::Exceptions::LogFile->throw("Error seeking to position: $!\n");
+		}
+		
+#		print STDERR 'Reading buffer for line '. $current_line .".\n" if $debug{refaddr $self};
+		
+		# Check if we've reached the end of the file.
+		# (And that we haven't gone back...)
+		if ( defined($parse_buffer{$$self}->[0]) 
+			and $#{$parse_buffer{$$self}} < $parse_buffer_size{$$self}
+			and $current_line >= $parse_buffer_start_line{$$self}
+			) {
+			return $parse_buffer{$$self}->[-1];
+		}
+		
+		# Clear the buffer.
+		@{$parse_buffer{$$self}} = ();
+		
+		# Read in the buffer.
+		READ_LOOP: for my $i (0...$parse_buffer_size{$$self}) {
+			$parse_buffer{$$self}->[$i] = $self->_parse_next_line();
+			last READ_LOOP unless defined $parse_buffer{$$self}->[$i];
+		}
+		
+#use Data::Dumper;
+#print STDERR Data::Dumper->Dump($parse_buffer{refaddr $self});
+		
+		# Move the indexes back to the line we are reading.
+		# (Note the 'current line' direct access again...)
+		$parse_buffer_start_line{$$self} = $current_line{$$self} - $#{$parse_buffer{$$self}};
+		$self->go_to_line_number($parse_buffer_start_line{$$self});
+		
+		# Return the data.
+		return $parse_buffer{$$self}->[0];
+	}
 }
 
 =head2 previous
@@ -275,7 +361,7 @@ sub previous {
 	my ($self) = @_;
 
 	# Check if we can.
-	if ( $log_info{refaddr $self}->{'current_line'} <= 1 ) {
+	if ( $current_line{$$self} <= 1 ) {
 		return undef;
 	}
 
@@ -300,24 +386,35 @@ Example:
 =cut
 
 sub go_forward {
-	my ($self, $lines) = @_;
+	my $self = shift;
+	my $lines = shift;
 	
 	# Just because I'm paranoid.
 	$lines ||= 1;
 	
 	# If we've read the line before, go straight to it.
-	if ( ${$log_info{refaddr $self}{'line_positions'}}[($log_info{refaddr $self}->{'current_line'}+$lines)] ) {
-		$log_info{refaddr $self}{'filehandle'}->setpos(${$log_info{refaddr $self}{'line_positions'}}[($log_info{refaddr $self}->{'current_line'}+$lines)])
-			or Mail::Log::Exceptions::LogFile->throw("Error seeking to position: $!\n");
-		$log_info{refaddr $self}->{'current_line'} = $log_info{refaddr $self}->{'current_line'} + $lines;
+	if ( ${$log_info{$$self}{line_positions}}[($current_line{$$self}+$lines)] ) {
+		$current_line{$$self} = $current_line{$$self} + $lines;
+		return 1;
 	}
 	else {
-		# Otherwise, read until we get to it.
-		foreach ( 1..$lines ) {
-			$self->next();
+		# Work out where we are.
+		my $start_pos = $self->get_line_number();
+		my $end_known_pos = $#{$log_info{$$self}{line_positions}};	# zero-indexed.
+		my $lines_remaining = $lines - ($end_known_pos - $start_pos);
+
+		# Go to the last line we have.
+		$current_line{$$self} = $#{$log_info{$$self}{line_positions}};
+		
+		# Then read until we get to the line we want.
+		if ( $self->next() ) {
+			unshift @_, ($self, $lines_remaining - 1 );
+			goto &go_forward;
+		}
+		else {
+			return 0;
 		}
 	}
-	return 1;
 }
 
 =head2 go_backward
@@ -343,16 +440,13 @@ sub go_backward {
 	$lines ||= 1;
 
 	# If the line exits, go straight to it.
-	if ( ($log_info{refaddr $self}->{'current_line'} - $lines ) > 0 ) {
-		$log_info{refaddr $self}{'filehandle'}->setpos(${$log_info{refaddr $self}{'line_positions'}}[($log_info{refaddr $self}->{'current_line'}-$lines)])
-			or Mail::Log::Exceptions::LogFile->throw("Error seeking to position: $!\n");
-		$log_info{refaddr $self}{'current_line'} -= $lines;
+	if ( ($current_line{$$self} - $lines ) > 0 ) {
+		$current_line{$$self} -= $lines;
 	}
 	else {
 		#If they've asked us to go beyond the beginning of the file, just go to the beginning.
-		$log_info{refaddr $self}{'filehandle'}->setpos(${$log_info{refaddr $self}{'line_positions'}}[0])
-			or Mail::Log::Exceptions::LogFile->throw("Error seeking to position: $!\n");
-		$log_info{refaddr $self}->{'current_line'} = 0;
+		$current_line{$$self} = 0;
+		return 0;
 	}
 	return 1;
 }
@@ -368,9 +462,7 @@ Returns true on success.
 sub go_to_beginning {
 	my ($self) = @_;
 
-	$log_info{refaddr $self}{'filehandle'}->setpos(${$log_info{refaddr $self}{'line_positions'}}[0])
-		or Mail::Log::Exceptions::LogFile->throw("Error seeking to beginning: $!\n");
-	$log_info{refaddr $self}->{'current_line'} = 0;
+	$current_line{$$self} = 0;
 
 	return 1;
 }
@@ -388,15 +480,15 @@ Returns true on success.
 sub go_to_end {
 	my ($self) = @_;
 	
-		$log_info{refaddr $self}{'filehandle'}->setpos(${$log_info{refaddr $self}{'line_positions'}}[-1])
-			or Mail::Log::Exceptions::LogFile->throw("Error seeking to end: $!\n");
-		$log_info{refaddr $self}->{'current_line'} = $#{$log_info{refaddr $self}{'line_positions'}};
+	# Go to the end of what we have.
+	$current_line{$$self} = $#{$log_info{$$self}{line_positions}};
 
-	while ( $self->next() ) {
-		1;
+	if ( !$self->next() ) {
+		return 1;
+	}
+	else {
+		goto &go_to_end;
 	}
-
-	return 1;
 }
 
 =head2 get_line_number
@@ -412,9 +504,40 @@ Example:
 
 =cut
 
-sub get_line_number {
+sub get_line_number () {
+	# This method gets called a lot: speed is an issue.
+	# This is as fast as I could make it.
+	return $current_line{${$_[0]}};
+}
+
+=head2 go_to_line_number 
+
+Goes to a specific logical line number.  (Preferably one that exits...)
+
+=cut
+
+sub go_to_line_number {
+	my ($self, $line_number) = @_;
+
+#	my $current_line_number = $self->get_line_number();
+
+	no warnings qw(uninitialized);
+	if ( $current_line{$$self} >= $line_number ) {
+		return $self->go_backward($current_line{$$self} - $line_number);
+	}
+	else {
+		return $self->go_forward($line_number - $current_line{$$self});
+	}
+}
+
+#
+# To be overrridden by subclasses.
+#
+
+sub _parse_next_line {
 	my ($self) = @_;
-	return $log_info{refaddr $self}{'current_line'};
+	
+	Mail::Log::Exceptions::Unimplemented->throw("Method '_parse_next_line' needs to be implemented by the subclass.\n");
 }
 
 #
@@ -441,8 +564,8 @@ C<_get_data_line>) is parsable in the currently understood format.
 sub _set_current_position_as_next_line { 
 	my ($self) = @_;
 
-	$log_info{refaddr $self}{'current_line'} += 1;
-	${$log_info{refaddr $self}{'line_positions'}}[$log_info{refaddr $self}->{'current_line'}] = $log_info{refaddr $self}{'filehandle'}->getpos()
+	$current_line{$$self} += 1;
+	${$log_info{$$self}{line_positions}}[$current_line{$$self}] = $log_info{$$self}{filehandle}->getpos()
 		or Mail::Log::Exceptions::LogFile->throw("Unable to get current file position: $!\n");
 	return;
 }
@@ -456,15 +579,23 @@ from the logfile, seperated by the current input seperator.
 
 sub _get_data_line {
 	my ($self) = @_;
-	return $log_info{refaddr $self}{'filehandle'}->getline();
+	return $log_info{$$self}{filehandle}->getline();
 }
 
+sub _clear_buffer {
+	my ($self) = @_;
+	@{$parse_buffer{$$self}} = undef;
+	$parse_buffer_start_line{$$self} = -1;
+	return;
+}
+
+
 =head2 Suggested usage:
 
-Suggestion on how to use the above two methods to implement a 'next' routine in
+Suggestion on how to use the above two methods to implement a '_parse_next_line' routine in
 a subclass:
 
-  sub next {
+  sub _parse_next_line {
 	my ($self) = @_;
 
 	# The hash we will return.
@@ -505,13 +636,19 @@ where you are in the file and what you've read so far.
 Those two methods should do slightly better on 'success' testing, to return
 better values.  (They basically always return true at the moment.)
 
+C<get_line_number> will return one less than the true line number if you are
+at the end of the file, and the buffer was completely filled.  (So that the
+end of the file is the last space of the buffer.)  Changing the buffer size or
+just going back and re-reading so that the buffer is restarted at a different
+location will allow you to retrieve the correct file length.
+
 =head1 REQUIRES
 
-Scalar::Util, File::Basename, IO::File, File::Temp, Mail::Log::Exceptions
+Scalar::Util, File::Basename, IO::File, Mail::Log::Exceptions
 
 =head1 RECOMMENDS
 
-IO::Uncompress::AnyUncompress
+IO::Uncompress::AnyUncompress, File::Temp
 
 =head1 AUTHOR
 
@@ -527,6 +664,16 @@ seeking through a file, both forwards and back.)
 
 =head1 HISTORY
 
+Dec 09, 2008 (1.2.10) - Profiled and sped up code.  (Cut processing time in half
+for some cases.)
+
+Nov 28, 2008 - Documentation fixes.
+
+Nov 18, 2008 - Now buffers reading, and prefers data from the buffer.
+
+Oct 24, 2008 - File::Temp now optional; only required for uncompressed files.
+               Added go_to_line_number for slightly better functionality.
+
 Oct 14, 2008 - Found that I need File::Temp of at least version 0.17.
 
 Oct 13, 2008 - Fixed tests so they do a better job of checking if they 
