@@ -29,14 +29,13 @@ use Mail::Log::Parse;
 use Mail::Log::Exceptions;
 use base qw(Mail::Log::Parse Exporter);
 
+use Memoize;
+memoize('timelocal');
+
 BEGIN {
     use Exporter ();
-    use vars qw($VERSION @EXPORT @EXPORT_OK %EXPORT_TAGS);
-    $VERSION     = '1.1.0';
-    #Give a hoot don't pollute, do not export more than needed by default
-    @EXPORT      = qw();
-    @EXPORT_OK   = qw();
-    %EXPORT_TAGS = ();
+    use vars qw($VERSION);
+    $VERSION     = '1.0410';
 }
 
 # A constant, to convert month names to month numbers.
@@ -72,11 +71,39 @@ my %log_info;
 sub DESTROY {
 	my ($self) = @_;
 	
-	delete $log_info{refaddr $self};
+	delete $log_info{$$self};
+	
+	$self->SUPER::DESTROY();
 	
 	return;
 }
 
+sub new {
+	my ($class, $parameters_ref) = @_;
+
+	my $self = $class->SUPER::new($parameters_ref);
+
+	if (defined($parameters_ref->{year})) {
+		$self->set_year($parameters_ref->{year});
+	}
+
+	return $self
+}
+
+=head2 set_year
+
+Sets the year, for the log timestamps.  If not set, the log is assumed to
+be for the current year.  (Can also be passed in C<new>, with the key 'year'.)
+
+=cut
+
+sub set_year {
+	my ($self, $year) = @_;
+	$log_info{refaddr $self}->{year} = $year;
+	$self->_clear_buffer();
+	return
+}
+
 =head2 next
 
 Returns a hash of the next line of postfix log data.
@@ -85,17 +112,19 @@ Hash keys are:
 
 	delay_before_queue, delay_connect_setup, delay_in_queue, 
 	delay_message_transmission, from, host, id, msgid, pid, program, 
-	relay, size, status, text, timestamp, to, total_delay
+	relay, size, status, text, timestamp, to, delay, connect,
+	disconnect
 
 All keys are garunteed to be present.  'program', 'pid', 'host', 'timestamp',
-'id' and 'text' are garunteed to have a value.
+'id' and 'text' are garunteed to have a value.  'connect' and 'disconnect' are
+boolean: true if the line is the relevant type of line, false otherwise.
 
 The 'text' key will have all of the log text B<after> the standard Postfix
 header.  (All of which is in the other keys that are required to have a value.)
 
 =cut
 
-sub next {
+sub _parse_next_line {
 	my ($self) = @_;
 
 	# The hash we will return.
@@ -124,7 +153,12 @@ sub next {
 	# First few fields are the date.  Convert back to Unix format...
 	{	# We don't need all these temp variables hanging around.
 		my ($log_hour, $log_minutes, $log_seconds) = split /:/, $line_data[2];
-		$line_info{timestamp} = timelocal($log_seconds, $log_minutes, $log_hour, $line_data[1], $MONTH_NUMBER{$line_data[0]}, $CURR_DATE[5]);
+		if (!defined($log_info{$$self}->{year}) ) {
+			$line_info{timestamp} = timelocal($log_seconds, $log_minutes, $log_hour, $line_data[1], $MONTH_NUMBER{$line_data[0]}, $CURR_DATE[5]);
+		}
+		else {
+			$line_info{timestamp} = timelocal($log_seconds, $log_minutes, $log_hour, $line_data[1], $MONTH_NUMBER{$line_data[0]}, $log_info{$$self}->{year});
+		}
 	}
 
 	# Machine Hostname
@@ -151,27 +185,33 @@ sub next {
 	# (These may or may not return any info...)
 
 	# To address
-	@{$line_info{to}} = $line_info{text} =~ m/to=([^,]*),/g;
+	@{$line_info{to}} = $line_info{text} =~ m/\bto=([^,]+),/g;
 
 	# From address
-	($line_info{from}) = $line_info{text} =~ m/from=([^,]*),/;
+	($line_info{from}) = $line_info{text} =~ m/\bfrom=([^,]+),/;
 
 	# Relay
-	($line_info{relay}) = $line_info{text} =~ m/relay=([^,]*),/;
+	($line_info{relay}) = $line_info{text} =~ m/\brelay=([^,]+),/;
 
 	# Status
-	($line_info{status}) = $line_info{text} =~ m/status=(.*)$/;
+	($line_info{status}) = $line_info{text} =~ m/\bstatus=(.+)$/;
 
 	# Size
-	($line_info{size}) = $line_info{text} =~ m/size=([^,]*),/;
+	($line_info{size}) = $line_info{text} =~ m/\bsize=([^,]+),/;
 
 	# Delays
 	($line_info{delay_before_queue}, $line_info{delay_in_queue}, $line_info{delay_connect_setup}, $line_info{delay_message_transmission} )
-		= $line_info{text} =~ m{delays=([\d.]+)/([\d.]+)/([\d.]+)/([\d.]+),};
-	($line_info{total_delay}) = $line_info{text} =~ m/delay=([\d.]+),/;
+		= $line_info{text} =~ m{\bdelays=([^/]+)/([^/]+)/([^/]+)/([^,]+),};
+	($line_info{delay}) = $line_info{text} =~ m/\bdelay=([\d.]+),/;
 
 	# Message ID
-	($line_info{msgid}) = $line_info{text} =~ m/message-id=(.*)$/;
+	($line_info{msgid}) = $line_info{text} =~ m/\bmessage-id=(.+)$/;
+
+	# Connect (Boolean)
+	$line_info{connect} = $line_info{text} =~ m/\bconnect from/;
+
+	# Disconnect (Boolean)
+	$line_info{disconnect} = $line_info{text} =~ m/\bdisconnect from/;
 
 	# Return the data.
 	return \%line_info;
@@ -183,7 +223,8 @@ None known at the moment.
 
 =head1 REQUIRES
 
-Scalar::Util, Time::Local, Mail::Log::Parse, Mail::Log::Exceptions
+L<Scalar::Util>, L<Time::Local>, L<Mail::Log::Parse>, L<Mail::Log::Exceptions>,
+L<Memoize>
 
 =head1 AUTHOR
 
@@ -197,6 +238,21 @@ L<Mail::Log::Parse>, for the main documentation on this module set.
 
 =head1 HISTORY
 
+Dec 09, 2008 (1.4.10) - Profiled code, did some speedups.  Added dependency on
+Memoize: For large logs this is a massive speedup.  For extremely sparse logs
+it may not be, but sparce logs are likely to be small.
+
+Nov 28, 2008 - Switched 'total_delay' to slightly more universal 'delay'.
+Sped up some regrexes.
+
+Nov 11, 2008 - Switched to using the bufferable C<_parse_next_line> instead of
+the unbuffered C<next>.
+
+Nov 6, 2008 - Added C<set_year> and alternate year handling, in case we aren't
+dealing with this year's logs.  (From the todo list.)
+
+Oct 24, 2008 - Added 'connect' and 'disconnect' members to the return hash.
+
 Oct 6, 2008 - Inital version.
 
 =head1 COPYRIGHT and LICENSE
